from page_scraper.core.node_utils import has_type
from page_scraper.entities.models import Entity, PageContext
from page_scraper.entities.builders import build_entity

class CategoryPageDetector:
    def detect(self, page: PageContext) -> list[Entity]:
        categories = []
        for entry  in page.nodes:
            node = entry['node']

            if not has_type(node, "CollectionPage"):
                continue

            categories.append(build_entity(node, "Category"))

        return categories


from page_scraper.core.node_utils import has_type
from page_scraper.entities.models import Entity, PageContext
from page_scraper.entities.builders import build_entity

class ProductDetector:
    def detect(self, page: PageContext) -> list[Entity]:
        products = []
        seen = set()

        for entry in page.nodes:
            node = entry['node']

            if not has_type(node, "Product"):
                continue

            entity = build_entity(node, "Product")

            key = (entity.url, (entity.name or "").lower())

            if key in seen:
                continue

            seen.add(key)

            products.append(entity)

        return products


from page_scraper.core.utils import process_breadcrumbs, infer_categories


def has_type(node: dict, type_name: str) -> bool:
    t = node.get("@type", [])

    if isinstance(t, str):
        return t == type_name
    return type_name in t

def has_types(node: dict, *type_names: str) -> bool:
    t = node.get("@type", [])
    if isinstance(t, str):
        return t in type_names

    return any(t_type in type_names for t_type in t)


def has_attrs(node: dict, *keys) -> bool:
    return any(node.get(key) for key in keys)

def path_contains(path: tuple, *keys) -> bool:
    return any(k in path for k in keys)


def has_bread_crumbs(node: dict):
    return has_type(node,"BreadcrumbList")

def get_categories(bread_crumbs:list[dict]) -> list[dict]:
    return infer_categories(bread_crumbs)

def build_entity_index(nodes: list[dict]) -> dict[str, dict]:
    index = {}
    for entry in nodes:
        node = entry["node"] if isinstance(entry, dict) and "node" in entry else entry
        node_id = node.get("@id")
        if node_id:
            index[node_id] = node
    return index

def extract_breadcrumbs(nodes:list) -> list:
    result = []
    for entry in nodes:
        node = entry["node"]
        if not has_bread_crumbs(node):
            continue
        return process_breadcrumbs(node)
        # result.append(cats)
    return result


import copy

from page_scraper.content.utils import get_clean_tags, process_media_tags, process_video_iframe, process_video_embed
from page_scraper.core.utils import canonical_domain
from page_scraper.entities.models import PageContext



class VideoTagScraper:
    def run(self,page:PageContext) -> PageContext:
        soup_copy = copy.copy(page.soup)

        clean = get_clean_tags(soup_copy,'video')
        domain = canonical_domain(page.url)
        videos = process_media_tags(clean,domain,'VIDEO_TAG')

        page.links.extend(videos)

        page.videos_count += len(videos)

        return page

class VideoIframeScraper:
    def run(self,page:PageContext) -> PageContext:
        soup_copy = copy.copy(page.soup)

        clean = get_clean_tags(soup_copy,'iframe')
        domain = canonical_domain(page.url)

        videos = process_video_iframe(clean,domain)

        page.videos_count += len(videos)

        return page

class VideoEmbedScraper:
    def run(self,page:PageContext) -> PageContext:
        soup_copy = copy.copy(page.soup)

        clean = get_clean_tags(soup_copy,'embed,object')
        domain = canonical_domain(page.url)
        videos = process_video_embed(clean,domain)

        page.links.extend(videos)

        page.videos_count += len(videos)

        return page




from page_scraper.core.node_utils import has_type
from page_scraper.entities.models import Entity, PageContext
from page_scraper.entities.builders import build_entity

class ContactPageDetector:
    def detect(self, page: PageContext) -> list[Entity]:
        contacts = []
        for entry  in page.nodes:
            node = entry['node']

            if not has_type(node, "ContactPage") and not has_type(node, "AboutPage"):
                continue

            contacts.append(build_entity(node, "Contact"))

        return contacts


from traceback import print_tb

import requests
from requests.exceptions import SSLError, ConnectionError, Timeout

class HttpFetcher:
    def __init__(self, user_agent: str | None = None, timeout: int = 20):
        self.session = requests.Session()
        self.timeout = timeout

        self.default_headers = {
            "User-Agent": user_agent or "Mozilla/5.0 (Linux; Android 15; SM-S931B) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/144.0.7559.96 Mobile Safari/537.36 (compatible; Googlebot/2.1; +http://www.google.com/bot.html)",
            "Accept": "text/html,application/xhtml+xml,application/xml;q=0.9,image/webp,*/*;q=0.8",
        }

    def get(self, url: str):
        try:
            response = self.session.get(
                url,
                headers=self.default_headers,
                timeout=self.timeout,
            )
            return response

        except SSLError:
            # retry without SSL verification
            response = self.session.get(
                url,
                headers=self.default_headers,
                timeout=self.timeout,
                verify=False,
            )
            return response

        except Timeout:
            return None  # or raise custom TimeoutError

        except ConnectionError:
            return None

    def head(self, url: str, allow_redirects=False):
        try:
            response = self.session.head(
                url,
                headers=self.default_headers,
                allow_redirects=allow_redirects,
                timeout=self.timeout,
            )
            return response

        except SSLError:
            # retry without SSL verification
            response = self.session.head(
                url,
                headers=self.default_headers,
                timeout=self.timeout,
                allow_redirects=allow_redirects,
                verify=False,
            )
            return response

        except Timeout:
            return None  # or raise custom TimeoutError

        except ConnectionError:
            return None



import copy

from page_scraper.content.utils import get_clean_tags, process_media_tags
from page_scraper.core.utils import canonical_domain
from page_scraper.entities.models import PageContext


class AudioTagScraper:
    def run(self,page:PageContext) -> PageContext:
        soup_copy = copy.copy(page.soup)

        clean = get_clean_tags(soup_copy,'audio')
        domain = canonical_domain(page.url)
        audio = process_media_tags(clean,domain,'AUDIO_TAG')

        page.links.extend(audio)

        page.audio_count += len(audio)

        return page


import copy

from page_scraper.entities.models import PageContext
from page_scraper.content.utils import get_clean_tags,process_headings


class HeadingScraper:
    def run(self,page:PageContext) -> PageContext:
        soup_copy = copy.copy(page.soup)

        headings = get_clean_tags(soup_copy,"h1,h2,h3,h4,h5,h6")
        page.headings = process_headings(headings)
        return page


import copy

from page_scraper.content.utils import get_clean_tags, process_image
from page_scraper.core.utils import canonical_domain
from page_scraper.entities.models import PageContext


class ImageScraper:
    def run(self,page:PageContext) -> PageContext:
        soup_copy = copy.copy(page.soup)

        clean = get_clean_tags(soup_copy,'img')
        domain = canonical_domain(page.url)
        images = process_image(clean,domain)

        page.links.extend(images)

        page.images_count = len(images)

        return page


def extract_nodes(data, parent=None, path=()):
    nodes = []

    if isinstance(data, list):
        for i, item in enumerate(data):
            nodes.extend(extract_nodes(item, parent, path + (i,)))

    elif isinstance(data, dict):
        if "@type" in data:
            nodes.append({
                "node": data,
                "parent": parent,
                "path": path
            })

        for key, value in data.items():
            nodes.extend(extract_nodes(value, data, path + (key,)))

    return nodes


import copy

from page_scraper.entities.models import PageContext
from page_scraper.meta.utils import get_meta_by_name


class RobotScraper:
    def run(self,page:PageContext) -> PageContext:
        soup_copy = copy.copy(page.soup)
        robots = get_meta_by_name(soup_copy, 'robots')[0].get('content','')

        page.is_index = False if 'noindex' in robots else True
        page.is_follow = False if 'nofollow' in robots else True

        return page


import copy
from page_scraper.entities.builders import build_response_meta
from page_scraper.entities.models import ResponseMeta, UrlContext

from page_scraper.core.utils import exclude_tags, process_text, canonical_url, process_link, classify_url

from bs4 import Tag

from page_scraper.infrastructure.fetcher import HttpFetcher

from requests import Response
from bs4 import BeautifulSoup

TAGS_TO_EXCLUDE = ("[class*=comment], [id*=comment], [id*=sidebar], [id*=footer], [id*=header],"
                   "nav, footer, aside, menu, sidebar, head, header, script")

fetcher = HttpFetcher()

def prepare_response_header(response: Response) -> ResponseMeta:
    return build_response_meta(response.headers)

def get_nav_links(soup: BeautifulSoup) -> list[Tag]:
    return [a for a in soup.select("nav a") if a.get("href")]

def get_clean_link(soup: BeautifulSoup) -> list[Tag]:
    cleaned = copy.copy(soup)
    exclude_tags(cleaned, TAGS_TO_EXCLUDE)
    return [
        a for a in cleaned.select("a")
        if a.get("href")
    ]



def prepare_links(domain:str,soup: list[Tag], position:str=None) -> list[UrlContext]:
    result = []
    for url in soup:
        can_url = canonical_url(url.get("href"), domain)
        if not can_url:
            continue

        classified_url = classify_url(can_url)

        if classified_url=="web":
            response = fetcher.head(can_url)
            data = process_link(can_url, response = response)
        else:
            data = process_link(can_url, url_type = classified_url)

        if not data:
            continue

        data.text = process_text(url)

        data.position=position

        result.append(data)

    return result


def get_page_canonical_url(soup:BeautifulSoup) -> Tag | None:
    return soup.find('link', rel="canonical")



from requests.structures import CaseInsensitiveDict

from page_scraper.entities.models import Entity, PageContext, ResponseMeta, PageMetaContext


def build_entity(node: dict, entity_type: str) -> Entity:
    return Entity(
        id=node.get("@id") or node.get("url"),
        type=entity_type,
        name=node.get("name"),
        alternate_name=node.get("alternateName"),
        url=node.get("url"),
        description=node.get("description"),
        raw=node,
        date_published=node.get("datePublished"),
        date_modified=node.get("dateModified"),
    )

def build_response_meta(meta:CaseInsensitiveDict) -> ResponseMeta:
    return ResponseMeta(
        content_length= meta.get("content-length"),
        content_encoding= meta.get("content-encoding"),
        cache_control= meta.get("cache-control"),
        last_modified= meta.get("last-modified"),
        etag= meta.get("etag"),
        server= meta.get("server"),
    )

def build_page_context(parsed_page:dict) -> PageContext:
    return PageContext(**parsed_page)

def build_page_meta(name, meta_type, value):
    return PageMetaContext(
        name=name,
        meta_type=meta_type,
        value=value,
    )




from page_scraper.content.base import ContentStep
from page_scraper.detectors.protocol import EntityDetector, PageStep
from page_scraper.entities.models import PageContext
from page_scraper.infrastructure.fetcher import HttpFetcher
from page_scraper.meta.base import MetaStep
from page_scraper.links.base import ScraperStep


class EntityPipeline:
    def __init__(self, detectors: list[EntityDetector]):
        self.detectors = detectors

    def run(self, page):
        entities = []
        for detector in self.detectors:
            entities.extend(detector.detect(page))
        page.entities = entities
        return entities

class CrawlerPipeline:
    def __init__(self, steps: list[PageStep]):
        self.steps = steps
    def process(self, url:str) -> PageContext:
        page = PageContext(url=url)
        for step in self.steps:
            page = step.run(page)
        return page

class ScraperPipeline:
    def __init__(self, steps: list[ScraperStep]):
        self.steps = steps

    def run(self, page: PageContext) -> PageContext:
        for step in self.steps:
            page = step.run(page)
        return page

class MetaPipeline:
    def __init__(self, steps: list[MetaStep]):
        self.steps = steps

    def run(self, page: PageContext) -> PageContext:
        for step in self.steps:
            page = step.run(page)
        return page

class ContentPipeline:
    def __init__(self, steps: list[ContentStep]):
        self.steps = steps

    def run(self, page: PageContext) -> PageContext:
        for step in self.steps:
            page = step.run(page)
        return page


import copy

from page_scraper.entities.models import PageContext
from page_scraper.links.utils import prepare_links, get_page_canonical_url
from page_scraper.core.utils import canonical_domain


class CanonicalScraper:
    def run(self,page:PageContext) -> PageContext:
        soup_copy = copy.copy(page.soup)

        url = get_page_canonical_url(soup_copy)

        if not url:
            return page

        domain = canonical_domain(page.url)
        result = prepare_links(domain, [url],"canonical")
        page.canonical_url = result[0]
        page.links.extend(result)

        return page


from bs4 import BeautifulSoup,Tag

from urllib.parse import urlparse, urlsplit, urlunsplit, unquote, quote, urljoin

from page_scraper.entities.models import UrlContext

import requests

SAFE_PATH="/-_.~@"

CONTENT_TYPE_MAP = {
    "application/ld+json": "json",
    "application/json": "json",
    "application/pdf": "pdf",
    "text/css": "stylesheet",
    "text/javascript": "javascript",
    "text/html": "html",
    "text/xml": "xml",
}

def is_root_url(url:str) -> bool:
    parsed = urlparse(url)
    return parsed.path in ("", "/")

def is_domain(node: dict) -> bool:
    url = node.get("url")
    return bool(url) and is_root_url(url)

def process_breadcrumbs(node:dict):
    breadcrumbs = []
    for li in node.get("itemListElement", []):
        item = li.get("item")
        crumb = {
            "position": li.get("position"),
            "name": li.get("name") or (item.get("name") if isinstance(item, dict) else None),
            "url": item.get("@id") if isinstance(item, dict) else item,
        }
        breadcrumbs.append(crumb)

    # sort explicitly (VERY IMPORTANT)
    breadcrumbs.sort(key=lambda x: x["position"] or 0)
    return breadcrumbs


def infer_categories(breadcrumbs):
    if len(breadcrumbs) <= 1:
        return []

    drop_first = is_domain(breadcrumbs[0])

    start = 1 if drop_first else 0

    return breadcrumbs[start:-1]

def clear_text(text:str) -> str | None:
    return text.strip().replace("\n", " ") or None

def process_text(tag:Tag)->str:
    return clear_text(tag.get_text())

def canonical_url(url: str, base:str) -> str:
    url = urljoin(base,url)
    parts = urlsplit(url)

    path = unquote(parts.path)
    path = quote(path, safe=SAFE_PATH)

    query = unquote(parts.query)
    query = quote(query, safe="=&")

    return urlunsplit((
        parts.scheme.lower(),
        parts.netloc.lower(),
        path,
        query,
        ""  # drop fragment for canonical URLs
    ))

def canonical_domain(url: str) -> str:
    split_url = urlsplit(url)
    return f"{split_url.scheme}://{split_url.netloc}"

def exclude_tags(soup: BeautifulSoup, selector: str) -> None:
    for tag in soup.select(selector):
        tag.decompose()

def get_page_type(content_type: str | None) -> str | None:
    if not content_type:
        return None

    ct = content_type.lower()

    # exact / substring matches first
    for key, value in CONTENT_TYPE_MAP.items():
        if key in ct:
            return value

    # prefix matches
    if ct.startswith("audio/"):
        return "podcast"
    if ct.startswith("image/"):
        return "image"
    if ct.startswith("video/"):
        return "video"

    return None

def process_link(current:str,response: requests.Response | None = None, url_type:str | None = None) -> UrlContext:
    status_code = 0
    content_type = url_type
    if response:
        status_code = response.status_code
        content_type = get_page_type(response.headers.get("content-type")),

    return UrlContext(
        url=current,
        status_code=status_code,
        content_type=content_type,
    )

def classify_url(url: str) -> str:
    parsed = urlsplit(url)

    if parsed.scheme == "mailto":
        return "email"
    if parsed.scheme in {"tel", "sms"}:
        return "phone"
    if parsed.scheme in {"whatsapp", "skype"}:
        return "messaging"
    if not parsed.scheme or parsed.scheme in ("http", "https"):
        return "web"

    return "other"


from page_scraper.entities.models import PageContext
from page_scraper.links.utils import get_nav_links, prepare_links
from page_scraper.core.utils import canonical_domain


class GetNavLinks:

    def run(self, page:PageContext) -> PageContext:
        soup = page.soup
        navigations = get_nav_links(soup)
        domain = canonical_domain(page.url)
        result = prepare_links(domain, navigations,"navigation")
        page.links.extend(result)
        return page
